from shapely.geometry import shape, Point, Polygon, mapping
import json

from asf_search import ASFSession, ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions 
from asf_search.download import download_url
from asf_search.CMR import translate_product
from remotezip import RemoteZip

from asf_search.download.file_download_type import FileDownloadType


class ASFProduct:
    def __init__(self, args: dict = {}, session: ASFSession = ASFSession()):
        self.meta = args.get('meta')
        self.umm = args.get('umm')

        translated = translate_product(args)

        self.properties = translated['properties']
        self.geometry = translated['geometry']
        self.baseline = translated['baseline']
        self.session = session

    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def geojson(self) -> dict:
        return {
            'type': 'Feature',
            'geometry': self.geometry,
            'properties': self.properties
        }

    def download(self, path: str, filename: str = None, session: ASFSession = None, fileType = FileDownloadType.DEFAULT_FILE) -> None:
        """
        Downloads this product to the specified path and optional filename.

        :param path: The directory into which this product should be downloaded.
        :param filename: Optional filename to use instead of the original filename of this product.
        :param session: The session to use, defaults to the one used to find the results.

        :return: None
        """
        if filename is None:
            default_filename = self.properties['fileName']
        
        if session is None:
            session = self.session

        urls = []

        def get_additional_urls():
            output = []
            base_filename = '.'.join(default_filename.split('.')[:-1])
            for url in self.properties['additionalUrls']:
                extension = url.split('.')[-1]
                urls.append((f"{base_filename}.{extension}", url))
            
            return output

        if fileType == FileDownloadType.DEFAULT_FILE:
            urls.append((default_filename, self.properties['url']))
        elif fileType == FileDownloadType.ADDITIONAL_FILES:
            urls.extend(get_additional_urls())
        elif fileType == FileDownloadType.ALL_FILES:
            urls.append((default_filename, self.properties['url']))
            urls.extend(get_additional_urls())
        else:
            raise ValueError("Invalid FileDownloadType provided, the valid types are 'DEFAULT_FILE', 'ADDITIONAL_FILES', and 'ALL_FILES'")

        for filename, url in urls:
            download_url(url=url, path=path, filename=filename, session=session)

    def stack(
            self,
            opts: ASFSearchOptions = None
    ) -> ASFSearchResults:
        """
        Builds a baseline stack from this product.

        :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

        :return: ASFSearchResults containing the stack, with the addition of baseline values (temporal, perpendicular) attached to each ASFProduct.
        """
        from .search.baseline_search import stack_from_product

        if opts is None:
            opts = ASFSearchOptions(session=self.session)

        return stack_from_product(self, opts=opts)

    def get_stack_opts(self) -> ASFSearchOptions:
        """
        Build search options that can be used to find an insar stack for this product

        :return: ASFSearchOptions describing appropriate options for building a stack from this product
        """
        from .search.baseline_search import get_stack_opts

        return get_stack_opts(reference=self)

    def centroid(self) -> Point:
        """
        Finds the centroid of a product
        """
        coords = mapping(shape(self.geometry))['coordinates'][0]
        lons = [p[0] for p in coords]
        if max(lons) - min(lons) > 180:
            unwrapped_coords = [a if a[0] > 0 else [a[0] + 360, a[1]] for a in coords]
        else:
            unwrapped_coords = [a for a in coords]

        return Polygon(unwrapped_coords).centroid

    def remotezip(self, session: ASFSession) -> RemoteZip:
        """Returns a RemoteZip object which can be used to download a part of an ASFProduct's zip archive.
        (See example in examples/5-Download.ipynb)
        
        :param session: an authenticated ASFSession
        """
        from .download.download import remotezip

        return remotezip(self.properties['url'], session=session)



from collections import UserList
from multiprocessing import Pool
import json
from asf_search import ASFSession, ASFSearchOptions
from asf_search.download.file_download_type import FileDownloadType
from asf_search.exceptions import ASFSearchError

from asf_search import ASF_LOGGER
from asf_search.export.csv import results_to_csv
from asf_search.export.jsonlite import results_to_jsonlite
from asf_search.export.jsonlite2 import results_to_jsonlite2
from asf_search.export.kml import results_to_kml
from asf_search.export.metalink import results_to_metalink

class ASFSearchResults(UserList):
    def __init__(self, *args, opts: ASFSearchOptions = None):
        super().__init__(*args)
        # Store it JUST so the user can access it (There might be zero products)
        # Each product will use their own reference to opts (but points to the same obj)
        self.searchOptions = opts
        self.searchComplete = False

    def geojson(self):
        return {
            'type': 'FeatureCollection',
            'features': [product.geojson() for product in self]
        }

    def csv(self):
        return results_to_csv(self)

    def kml(self):
        return results_to_kml(self)
    
    def metalink(self):
        return results_to_metalink(self)

    def jsonlite(self):
        return results_to_jsonlite(self)

    def jsonlite2(self):
        return results_to_jsonlite2(self)  

    def __str__(self):
        return json.dumps(self.geojson(), indent=2, sort_keys=True)

    def download(
            self,
            path: str,
            session: ASFSession = None,
            processes: int = 1,
            fileType = FileDownloadType.DEFAULT_FILE
    ) -> None:
        """
        Iterates over each ASFProduct and downloads them to the specified path.

        :param path: The directory into which the products should be downloaded.
        :param session: The session to use. Defaults to the session used to fetch the results, or a new one if none was used.
        :param processes: Number of download processes to use. Defaults to 1 (i.e. sequential download)

        :return: None
        """
        ASF_LOGGER.info(f"Started downloading ASFSearchResults of size {len(self)}.")
        if processes == 1:
            for product in self:
                product.download(path=path, session=session, fileType=fileType)
        else:
            ASF_LOGGER.info(f"Using {processes} threads - starting up pool.")
            pool = Pool(processes=processes)
            args = [(product, path, session, fileType) for product in self]
            pool.map(_download_product, args)
            pool.close()
            pool.join()
        ASF_LOGGER.info(f"Finished downloading ASFSearchResults of size {len(self)}.")
        
    def raise_if_incomplete(self) -> None:
        if not self.searchComplete:
            msg = "Results are incomplete due to a search error. See logging for more details. (ASFSearchResults.raise_if_incomplete called)"
            ASF_LOGGER.error(msg)
            raise ASFSearchError(msg)

def _download_product(args) -> None:
    product, path, session, fileType = args
    product.download(path=path, session=session, fileType=fileType)



import numbers
from typing import List
from asf_search.ASFSession import ASFSession
from requests.cookies import create_cookie
import http.cookiejar
import requests

from unittest.mock import patch

def run_auth_with_creds(username: str, password: str):
    session = ASFSession()
    session.auth_with_creds(username=username, password=password)

def run_auth_with_token(token: str):
    session = ASFSession()

    with patch('asf_search.ASFSession.get') as mock_token_session:
        if not token.startswith('Bearer EDL'):
                mock_token_session.return_value.status_code = 400
                session.auth_with_token(token)

        mock_token_session.return_value.status_code = 200
        session.auth_with_token(token)

def run_auth_with_cookiejar(cookies: List):
    cookiejar = http.cookiejar.CookieJar()
    for cookie in cookies:
        cookiejar.set_cookie(create_cookie(name=cookie.pop('name'), **cookie))
    session = ASFSession()
    session.auth_with_cookiejar(cookies)

def run_test_asf_session_rebuild_auth(
    original_domain: str, 
    response_domain: str, 
    response_code: numbers.Number, 
    final_token
    ):

    if final_token == 'None':
        final_token = None

    session = ASFSession()

    with patch('asf_search.ASFSession.get') as mock_token_session:
        mock_token_session.return_value.status_code = 200
        session.auth_with_token("bad_token")
        
        req = requests.Request(original_domain)
        req.headers.update({'Authorization': 'Bearer fakeToken'})

        response = requests.Response()
        response.status_code = response_code
        response.location = response_domain
        response.request = requests.Request()
        response.request.url = response_domain
        response.headers.update({'Authorization': 'Bearer fakeToken'})

        with patch('asf_search.ASFSession._get_domain') as hostname_patch:
            hostname_patch.side_effect = [original_domain, response_domain]
      
            session.rebuild_auth(req, response)

            assert req.headers.get("Authorization") == final_token


from .export_translators import ASFSearchResults_to_properties_list
from .csv import results_to_csv
from .metalink import results_to_metalink
from .kml import results_to_kml
from .jsonlite import results_to_jsonlite
from .jsonlite2 import results_to_jsonlite2
from .geojson import results_to_geojson



# backport of importlib.metadata for python < 3.8
from importlib_metadata import PackageNotFoundError, version
import logging

try:
    __version__ = version(__name__)
except PackageNotFoundError as e:
    msg = str('package is not installed!\n'
          'Install in editable/develop mode via (from the top of this repo):\n'
          '   python3 -m pip install -e .\n'
          'Or, to just get the version number use:\n'
          '   python setup.py --version')
    print(msg)
    ASF_LOGGER.exception(msg)
    raise PackageNotFoundError("Install with 'python3 -m pip install -e .' to use") from e

ASF_LOGGER = logging.getLogger(__name__)
# Add null handle so we do nothing by default. It's up to whatever
# imports us, if they want logging.
ASF_LOGGER.addHandler(logging.NullHandler())

from .ASFSession import ASFSession
from .ASFProduct import ASFProduct
from .ASFSearchResults import ASFSearchResults
from .ASFSearchOptions import ASFSearchOptions, validators
from .exceptions import *
from .constants import *
from .health import *
from .search import *
from .download import *
from .CMR import *
from .baseline import *
from .WKT import validate_wkt
from .export import *

REPORT_ERRORS=True 
"""Enables automatic search error reporting to ASF, send any questions to uso@asf.alaska.edu"""



from typing import Dict
from asf_search import ASFSearchOptions, ERROR_REPORTING_ENDPOINT
import requests
import logging

def report_search_error(search_options: ASFSearchOptions, message: str):
    """Reports CMR Errors automatically to ASF"""

    from asf_search import REPORT_ERRORS

    if not REPORT_ERRORS:
        logging.warning("Automatic search error reporting is turned off, search errors will NOT be reported to ASF.\
            \nTo enable automatic error reporting, set asf_search.REPORT_ERRORS to True\
            \nIf you have any questions email uso@asf.alaska.edu")
        return

    user_agent = search_options.session.headers.get("User-Agent")
    search_options_list = '\n'.join([f"\t{option}: {key}" for option, key in dict(search_options).items()])
    message=f"Error Message: {str(message)}\nUser Agent: {user_agent} \
    \nSearch Options: {{\n{search_options_list}\n}}"

    response = requests.post(f'https://{ERROR_REPORTING_ENDPOINT}', data={'Message': "This error message and info was automatically generated:\n\n" + message})

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        logging.error(f"asf-search failed to automatically report an error, if you have any questions email uso@asf.alaska.edu\
            \nError Text: HTTP {response.status_code}: {response.json()['errors']}")
        return
    if response.status_code == 200:
        logging.error("The asf-search module ecountered an error with CMR, and the following message was automatically reported to ASF:\n\n" + 
                        "\"\n" +
                        message + 
                        "\n\"" +
                        "If you have any questions email uso@asf.alaska.edu")



from asf_search.search.search_generator import *
from asf_search import ASFSearchOptions, ASFSearchResults
from asf_search import INTERNAL
from typing import List

import math

def run_test_search_generator_multi(search_opts: List[ASFSearchOptions]):
    queries = [search_generator(opts=opts) for opts in search_opts]
    
    expected_results_size = sum([opts.maxResults for opts in search_opts])
    expected_page_count = sum([math.ceil(opts.maxResults / INTERNAL.CMR_PAGE_SIZE) for opts in search_opts])
    combined_results = []
    
    page_count = 0
    searches = {}
    
    for opt in search_opts:
        if isinstance(opt.platform, list):
            for platform in opt.platform:
                searches[platform] = False
        else:
            searches[opt.platform] = False

    while(len(queries)):
        queries_iter = iter(queries)
        for idx, query in enumerate(queries_iter):  # Alternate pages between results
            page = next(query, None)
            if page != None:
                combined_results.extend(page)
                page_count += 1
                if page.searchComplete:
                    if isinstance(page.searchOptions.platform, list):
                        for platform in page.searchOptions.platform:
                            searches[platform] = True
                    else:
                        searches[page.searchOptions.platform] = True
            else:
                queries[idx] = None

        queries = [query for query in queries if query != None]

    assert page_count == expected_page_count
    assert len(combined_results) == expected_results_size
    assert len([completed for completed in searches if completed]) >= len(search_opts)

def run_test_search_generator(search_opts: ASFSearchOptions):
    pages_iter = search_generator(opts=search_opts)
    
    page_count = int(search_opts.maxResults / INTERNAL.CMR_PAGE_SIZE)

    page_idx = 0
    
    results = ASFSearchResults([])
    for page in pages_iter:
        results.extend(page)
        results.searchComplete = page.searchComplete
        results.searchOptions = page.searchOptions
        page_idx += 1
        
    assert page_count <= page_idx
    assert len(results) <= search_opts.maxResults
    assert results.searchComplete == True
    
    preprocess_opts(search_opts)

    for key, val in search_opts:
        if key != 'maxResults':
            assert getattr(results.searchOptions, key) == val


from numbers import Number
from asf_search import ASFSearchOptions
from asf_search.ASFProduct import ASFProduct
from asf_search.constants import INTERNAL
from asf_search.exceptions import ASFSearchError
from asf_search.search import search
from asf_search.ASFSearchResults import ASFSearchResults

import requests

from pytest import raises

import requests_mock

def run_test_ASFSearchResults(search_resp):
    search_results = ASFSearchResults([ASFProduct(product) for product in search_resp])

    assert(len(search_results) == len(search_resp))
    assert(search_results.geojson()['type'] == 'FeatureCollection')

    for (idx, feature) in enumerate(search_results):
        # temporal and perpendicular baseline values are calculated post-search, 
        # so there's no instance where they'll be returned in a CMR search
        search_resp[idx]['properties'].pop('temporalBaseline', None)
        search_resp[idx]['properties'].pop('perpendicularBaseline', None)

        assert(feature.geojson()['geometry'] == search_resp[idx]['geometry'])
        for key, item in feature.geojson()['properties'].items():
            assert(item == search_resp[idx]['properties'][key])

def run_test_search(search_parameters, answer):
    with requests_mock.Mocker() as m:
        m.post(f"https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}", json={'items': answer})
        response = search(**search_parameters)

        if search_parameters.get("maxResults", False):
            assert(len(response) == search_parameters["maxResults"])

        assert(len(response) == len(answer))
        # assert(response.geojson()["features"] == answer)

def run_test_search_http_error(search_parameters, status_code: Number, report: str):
    
    if not len(search_parameters.keys()):
        with requests_mock.Mocker() as m:
            m.register_uri('POST', f"https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}", status_code=status_code, json={'errors': {'report': report}})
            m.register_uri('POST', f"https://search-error-report.asf.alaska.edu/", real_http=True)
            searchOptions = ASFSearchOptions(**search_parameters)
            results = search(opts=searchOptions)
            assert len(results) == 0
            with raises(ASFSearchError):
                results.raise_if_incomplete()
            return

    # If we're not doing an empty search we want to fire off one real query to CMR, then interrupt it with an error
    # We can tell a search isn't the first one by checking if 'CMR-Search-After' has been set 
    def custom_matcher(request: requests.Request):
        if 'CMR-Search-After' in request.headers.keys():
            resp = requests.Response()
            resp.status_code = 200
            return resp
        return None

    with requests_mock.Mocker() as m:
        m.register_uri('POST', f"https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}", real_http=True)
        m.register_uri('POST', f"https://{INTERNAL.CMR_HOST}{INTERNAL.CMR_GRANULE_PATH}", additional_matcher=custom_matcher, status_code=status_code, json={'errors': {'report': report}})
        m.register_uri('POST', f"https://search-error-report.asf.alaska.edu/", real_http=True)

        search_parameters['maxResults'] = INTERNAL.CMR_PAGE_SIZE + 1
        searchOptions = ASFSearchOptions(**search_parameters)
        results = search(opts=searchOptions)
        
        assert results is not None
        assert 0 < len(results) <= INTERNAL.CMR_PAGE_SIZE
        with raises(ASFSearchError):
            results.raise_if_incomplete()

            



from asf_search.baseline.stack import get_baseline_from_stack, get_default_product_type
from copy import copy

from asf_search.search import search, product_search
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFProduct import ASFProduct
from asf_search.constants import PLATFORM
from asf_search.exceptions import ASFSearchError, ASFBaselineError
from copy import copy

precalc_platforms = [
    PLATFORM.ALOS,
    PLATFORM.RADARSAT,
    PLATFORM.ERS1,
    PLATFORM.ERS2,
    PLATFORM.JERS,
]


def stack_from_product(
        reference: ASFProduct,
        opts: ASFSearchOptions = None
    ) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference ASFProduct

    :param reference: Reference scene to base the stack on, and from which to calculate perpendicular/temporal baselines
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(dict) of search results
    """

    opts = (ASFSearchOptions() if opts is None else copy(opts))

    stack_opts = get_stack_opts(reference, opts=opts)

    stack = search(opts=stack_opts)
    is_complete = stack.searchComplete

    stack, warnings = get_baseline_from_stack(reference=reference, stack=stack)
    stack.searchComplete = is_complete # preserve final outcome of earlier search()

    stack.sort(key=lambda product: product.properties['temporalBaseline'])

    return stack

def stack_from_id(
        reference_id: str,
        opts: ASFSearchOptions = None
) -> ASFSearchResults:
    """
    Finds a baseline stack from a reference product ID

    :param reference_id: Reference product to base the stack from, and from which to calculate perpendicular/temporal baselines
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """

    opts = (ASFSearchOptions() if opts is None else copy(opts))

    reference_results = product_search(product_list=reference_id, opts=opts)

    if len(reference_results) <= 0:
        raise ASFSearchError(f'Reference product not found: {reference_id}')
    reference = reference_results[0]

    return stack_from_product(reference, opts=opts)


def get_stack_opts(
        reference: ASFProduct,
        opts: ASFSearchOptions = None
) -> ASFSearchOptions:

    stack_opts = (ASFSearchOptions() if opts is None else copy(opts))
    stack_opts.processingLevel = get_default_product_type(reference)

    if reference.properties['platform'] in precalc_platforms:
        if reference.properties['insarStackId'] not in [None, 'NA', 0, '0']:
            stack_opts.insarStackId = reference.properties['insarStackId']
            return stack_opts
        raise ASFBaselineError(f'Requested reference product needs a baseline stack ID but does not have one: {reference.properties["fileID"]}')

    # build a stack from scratch if it's a non-precalc dataset with state vectors
    
    if reference.properties['processingLevel'] == 'BURST':
        stack_opts.fullBurstID = reference.properties['burst']['fullBurstID']
        stack_opts.polarization = [reference.properties['polarization']]
        return stack_opts
    elif reference.properties['platform'].upper() in [PLATFORM.SENTINEL1A.upper(), PLATFORM.SENTINEL1B.upper()]:
        stack_opts.platform = [PLATFORM.SENTINEL1A, PLATFORM.SENTINEL1B]
        
        stack_opts.beamMode = [reference.properties['beamModeType']]
        stack_opts.flightDirection = reference.properties['flightDirection']
        stack_opts.relativeOrbit = [int(reference.properties['pathNumber'])]  # path
        
        if reference.properties['polarization'] in ['HH', 'HH+HV']:
            stack_opts.polarization = ['HH','HH+HV']
        elif reference.properties['polarization'] in ['VV', 'VV+VH']:
            stack_opts.polarization = ['VV','VV+VH']
        else:
            stack_opts.polarization = [reference.properties['polarization']]
        
        stack_opts.intersectsWith = reference.centroid().wkt
        
        return stack_opts

    raise ASFBaselineError(f'Reference product is not a pre-calculated baseline dataset, and not a known ephemeris-based dataset: {reference.properties["fileID"]}')



from typing import Union, Iterable
import datetime
from copy import copy

from asf_search.search import search
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults


def geo_search(
        intersectsWith: str,
        absoluteOrbit: Iterable[Union[int, range]] = None,
        asfFrame: Iterable[Union[int, range]] = None,
        beamMode: Iterable[str] = None,
        beamSwath: Union[str, Iterable[str]] = None,
        campaign: Union[str, Iterable[str]] = None,
        end: Union[datetime.datetime, str] = None,
        flightDirection: Iterable[str] = None,
        frame: Iterable[Union[int, range]] = None,
        instrument: Iterable[str] = None,
        lookDirection: Iterable[str] = None,
        platform: Iterable[str] = None,
        polarization: Iterable[str] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Iterable[str] = None,
        relativeOrbit: Iterable[Union[int, range]] = None,
        start: Union[datetime.datetime, str] = None,
        maxResults: int = None,
        opts: ASFSearchOptions = None
) -> ASFSearchResults:
    """
    Performs a geographic search using the ASF SearchAPI

    :param absoluteOrbit: For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B this value corresponds to the orbit count within the orbit cycle. For UAVSAR it is the Flight ID.
    :param asfFrame: This is primarily an ASF / JAXA frame reference. However, some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    :param beamMode: The beam mode used to acquire the data.
    :param beamSwath: Encompasses a look angle and beam mode.
    :param campaign: For UAVSAR and AIRSAR data collections only. Search by general location, site description, or data grouping as supplied by flight agency or project.
    :param end: End date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param flightDirection: Satellite orbit direction during data acquisition
    :param frame: ESA-referenced frames are offered to give users a universal framing convention. Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    :param instrument: The instrument used to acquire the data. See also: platform
    :param intersectsWith: Search by polygon, linestring, or point defined in 2D Well-Known Text (WKT)
    :param lookDirection: Left or right look direction during data acquisition
    :param platform: Remote sensing platform that acquired the data. Platforms that work together, such as Sentinel-1A/1B and ERS-1/2 have multi-platform aliases available. See also: instrument
    :param polarization: A property of SAR electromagnetic waves that can be used to extract meaningful information about surface properties of the earth.
    :param processingDate: Used to find data that has been processed at ASF since a given time and date.  Supports timestamps as well as natural language such as "3 weeks ago"
    :param processingLevel: Level to which the data has been processed
    :param relativeOrbit: Path or track of satellite during data acquisition. For UAVSAR it is the Line ID.
    :param start: Start date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param maxResults: The maximum number of results to be returned by the search
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """

    kwargs = locals()
    data = dict((k, v) for k, v in kwargs.items() if k not in ['host', 'opts'] and v is not None)

    opts = (ASFSearchOptions() if opts is None else copy(opts))
    opts.merge_args(**data)

    return search(opts=opts)



from typing import Union, Iterable, Tuple
from copy import copy
import datetime

from asf_search import ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.search.search_generator import search_generator

def search(
        absoluteOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        asfFrame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        beamMode: Union[str, Iterable[str]] = None,
        beamSwath: Union[str, Iterable[str]] = None,
        campaign: Union[str, Iterable[str]] = None,
        maxDoppler: float = None,
        minDoppler: float = None,
        end: Union[datetime.datetime, str] = None,
        maxFaradayRotation: float = None,
        minFaradayRotation: float = None,
        flightDirection: str = None,
        flightLine: str = None,
        frame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        granule_list: Union[str, Iterable[str]] = None,
        groupID: Union[str, Iterable[str]] = None,
        insarStackId: str = None,
        instrument: Union[str, Iterable[str]] = None,
        intersectsWith: str = None,
        lookDirection: Union[str, Iterable[str]] = None,
        offNadirAngle: Union[float, Tuple[float, float], Iterable[Union[float, Tuple[float, float]]]] = None,
        platform: Union[str, Iterable[str]] = None,
        polarization: Union[str, Iterable[str]] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Union[str, Iterable[str]] = None,
        product_list: Union[str, Iterable[str]] = None,
        relativeOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        season: Tuple[int, int] = None,
        start: Union[datetime.datetime, str] = None,
        absoluteBurstID: Union[int, Iterable[int]] = None,
        relativeBurstID: Union[int, Iterable[int]] = None,
        fullBurstID: Union[str, Iterable[str]] = None,
        collections: Union[str, Iterable[str]] = None,
        temporalBaselineDays: Union[str, Iterable[str]] = None,
        maxResults: int = None,
        opts: ASFSearchOptions = None,
) -> ASFSearchResults:
    """
    Performs a generic search using the ASF SearchAPI. Accepts a number of search parameters, and/or an ASFSearchOptions object. If an ASFSearchOptions object is provided as well as other specific parameters, the two sets of options will be merged, preferring the specific keyword arguments.

    :param absoluteOrbit: For ALOS, ERS-1, ERS-2, JERS-1, and RADARSAT-1, Sentinel-1A, Sentinel-1B this value corresponds to the orbit count within the orbit cycle. For UAVSAR it is the Flight ID.
    :param asfFrame: This is primarily an ASF / JAXA frame reference. However, some platforms use other conventions. See ‘frame’ for ESA-centric frame searches.
    :param beamMode: The beam mode used to acquire the data.
    :param beamSwath: Encompasses a look angle and beam mode.
    :param campaign: For UAVSAR and AIRSAR data collections only. Search by general location, site description, or data grouping as supplied by flight agency or project.
    :param maxDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param minDoppler: Doppler provides an indication of how much the look direction deviates from the ideal perpendicular flight direction acquisition.
    :param end: End date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param maxFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param minFaradayRotation: Rotation of the polarization plane of the radar signal impacts imagery, as HH and HV signals become mixed.
    :param flightDirection: Satellite orbit direction during data acquisition
    :param flightLine: Specify a flightline for UAVSAR or AIRSAR.
    :param frame: ESA-referenced frames are offered to give users a universal framing convention. Each ESA frame has a corresponding ASF frame assigned. See also: asfframe
    :param granule_list: List of specific granules. Search results may include several products per granule name.
    :param groupID: Identifier used to find products considered to be of the same scene but having different granule names
    :param insarStackId: Identifier used to find products of the same InSAR stack
    :param instrument: The instrument used to acquire the data. See also: platform
    :param intersectsWith: Search by polygon, linestring, or point defined in 2D Well-Known Text (WKT)
    :param lookDirection: Left or right look direction during data acquisition
    :param offNadirAngle: Off-nadir angles for ALOS PALSAR
    :param platform: Remote sensing platform that acquired the data. Platforms that work together, such as Sentinel-1A/1B and ERS-1/2 have multi-platform aliases available. See also: instrument
    :param polarization: A property of SAR electromagnetic waves that can be used to extract meaningful information about surface properties of the earth.
    :param processingDate: Used to find data that has been processed at ASF since a given time and date. Supports timestamps as well as natural language such as "3 weeks ago"
    :param processingLevel: Level to which the data has been processed
    :param product_list: List of specific products. Guaranteed to be at most one product per product name.
    :param relativeOrbit: Path or track of satellite during data acquisition. For UAVSAR it is the Line ID.
    :param season: Start and end day of year for desired seasonal range. This option is used in conjunction with start/end to specify a seasonal range within an overall date range.
    :param start: Start date of data acquisition. Supports timestamps as well as natural language such as "3 weeks ago"
    :param collections: List of collections (concept-ids) to limit search to 
    :param temporalBaselineDays: List of temporal baselines, used for Sentinel-1 Interferogram (BETA)
    :param maxResults: The maximum number of results to be returned by the search
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """
    kwargs = locals()
    data = dict((k, v) for k, v in kwargs.items() if k not in ['host', 'opts'] and v is not None)

    opts = (ASFSearchOptions() if opts is None else copy(opts))
    opts.merge_args(**data)

    results = ASFSearchResults([])
    
    # The last page will be marked as complete if results sucessful
    for page in search_generator(opts=opts):
        results.extend(page)
        results.searchComplete = page.searchComplete
        results.searchOptions = page.searchOptions
    
    results.sort(key=lambda p: (p.properties['stopTime'], p.properties['fileID']), reverse=True)
    
    return results



field_map = {
    # API parameter               CMR keyword                       CMR format strings
    'absoluteOrbit':        {'key': 'orbit_number',            'fmt': '{0}'},
    'asfFrame':             {'key': 'attribute[]',             'fmt': 'int,FRAME_NUMBER,{0}'},
    'asfPlatform':          {'key': 'attribute[]',             'fmt': 'string,ASF_PLATFORM,{0}'},
    'maxBaselinePerp':      {'key': 'attribute[]',             'fmt': 'float,INSAR_BASELINE,,{0}'},
    'minBaselinePerp':      {'key': 'attribute[]',             'fmt': 'float,INSAR_BASELINE,{0},'},
    'bbox':                 {'key': 'bounding_box',            'fmt': '{0}'},
    'beamMode':             {'key': 'attribute[]',             'fmt': 'string,BEAM_MODE,{0}'},
    'beamSwath':            {'key': 'attribute[]',             'fmt': 'string,BEAM_MODE_TYPE,{0}'},
    'campaign':             {'key': 'attribute[]',             'fmt': 'string,MISSION_NAME,{0}'},
    'maxDoppler':           {'key': 'attribute[]',             'fmt': 'float,DOPPLER,,{0}'},
    'minDoppler':           {'key': 'attribute[]',             'fmt': 'float,DOPPLER,{0},'},
    'maxFaradayRotation':   {'key': 'attribute[]',             'fmt': 'float,FARADAY_ROTATION,,{0}'},
    'minFaradayRotation':   {'key': 'attribute[]',             'fmt': 'float,FARADAY_ROTATION,{0},'},
    'flightDirection':      {'key': 'attribute[]',             'fmt': 'string,ASCENDING_DESCENDING,{0}'},
    'flightLine':           {'key': 'attribute[]',             'fmt': 'string,FLIGHT_LINE,{0}'},
    'frame':                {'key': 'attribute[]',             'fmt': 'int,CENTER_ESA_FRAME,{0}'},
    'granule_list':         {'key': 'readable_granule_name[]', 'fmt': '{0}'},
    'groupID':              {'key': 'attribute[]',             'fmt': 'string,GROUP_ID,{0}'},
    'insarStackId':         {'key': 'attribute[]',             'fmt': 'int,INSAR_STACK_ID,{0}'},
    'linestring':           {'key': 'line',                    'fmt': '{0}'},
    'lookDirection':        {'key': 'attribute[]',             'fmt': 'string,LOOK_DIRECTION,{0}'},
    'maxInsarStackSize':    {'key': 'attribute[]',             'fmt': 'int,INSAR_STACK_SIZE,,{0}'},
    'minInsarStackSize':    {'key': 'attribute[]',             'fmt': 'int,INSAR_STACK_SIZE,{0},'},
    'instrument':           {'key': 'instrument[]',            'fmt': '{0}'},
    'offNadirAngle':        {'key': 'attribute[]',             'fmt': 'float,OFF_NADIR_ANGLE,{0}'},
    'platform':             {'key': 'platform[]',              'fmt': '{0}'},
    'polarization':         {'key': 'attribute[]',             'fmt': 'string,POLARIZATION,{0}'},
    'point':                {'key': 'point',                   'fmt': '{0}'},
    'polygon':              {'key': 'polygon',                 'fmt': '{0}'},
    'processingDate':       {'key': 'updated_since',           'fmt': '{0}'},
    'processingLevel':      {'key': 'attribute[]',             'fmt': 'string,PROCESSING_TYPE,{0}'},
    'product_list':         {'key': 'granule_ur[]',            'fmt': '{0}'},
    'provider':             {'key': 'provider',                'fmt': '{0}'},
    'relativeOrbit':        {'key': 'attribute[]',             'fmt': 'int,PATH_NUMBER,{0}'},
    'temporal':             {'key': 'temporal',                'fmt': '{0}'},
    'collections':          {'key': 'echo_collection_id[]',    'fmt': '{0}'},
    'temporalBaselineDays': {'key': 'attribute[]',             'fmt': 'int,TEMPORAL_BASELINE_DAYS,{0}'},
    
    # SLC BURST fields
    'absoluteBurstID':      {'key': 'attribute[]',             'fmt': 'int,BURST_ID_ABSOLUTE,{0}'},
    'relativeBurstID':      {'key': 'attribute[]',             'fmt': 'int,BURST_ID_RELATIVE,{0}'},
    'fullBurstID':          {'key': 'attribute[]',             'fmt': 'string,BURST_ID_FULL,{0}'},
}


import inspect
from types import GeneratorType
from asf_search import ASF_LOGGER
from asf_search.CMR import get_additional_fields
from asf_search.export.metalink import MetalinkStreamArray
import xml.etree.ElementTree as ETree

extra_kml_fields = [
    ('configurationName', ['AdditionalAttributes', ('Name', 'BEAM_MODE_DESC'), 'Values', 0]),
    ('faradayRotation', ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0]),
    ('processingTypeDisplay', ['AdditionalAttributes', ('Name', 'PROCESSING_TYPE_DISPLAY'), 'Values', 0]),
    ('sceneDate', ['AdditionalAttributes', ('Name', 'ACQUISITION_DATE'), 'Values', 0]),
    ('shape', ['SpatialExtent', 'HorizontalSpatialDomain', 'Geometry', 'GPolygons', 0, 'Boundary', 'Points']),
    ('thumbnailUrl', ['AdditionalAttributes', ('Name', 'THUMBNAIL_URL'), 'Values', 0]),
    ('faradayRotation', ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0]),
]

def results_to_kml(results):
    ASF_LOGGER.info('Started translating results to kml format')
    
    if inspect.isgeneratorfunction(results) or isinstance(results, GeneratorType):
        return KMLStreamArray(results)
    
    return KMLStreamArray([results])

class KMLStreamArray(MetalinkStreamArray):
    def __init__(self, results):
        MetalinkStreamArray.__init__(self, results)
        self.header = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
 <kml xmlns="http://www.opengis.net/kml/2.2">
   <Document>
     <name>ASF Datapool Search Results</name>
     <description>Search Performed:</description>
     <Style id="yellowLineGreenPoly">
         <LineStyle>
         <color>30ff8800</color>
         <width>4</width>
         </LineStyle>
         <PolyStyle>
         <color>7f00ff00</color>
         </PolyStyle>
     </Style>\n     """
        self.footer = """</Document>\n</kml>"""
        
    def get_additional_fields(self, product):
        umm = product.umm
        additional_fields = {}
        for key, path in extra_kml_fields:
            additional_fields[key] = get_additional_fields(umm, *path)
        return additional_fields

    def getOutputType(self) -> str:
        return 'kml'
    
    def getItem(self, p):
        placemark = ETree.Element("Placemark")
        name = ETree.Element('name')
        name.text = p['sceneName']
        placemark.append(name)
        
        description = ETree.Element('description')
        description.text = """&lt;![CDATA["""
        placemark.append(description)
        
        h1 = ETree.Element('h1')
        h1.text = f"{p['platform']} ({p['configurationName']}), acquired {p['sceneDate']}"
        h2 = ETree.Element('h2')
        h2.text = p.get('url', '')
        description.append(h1)
        description.append(h2)
        
        div = ETree.Element('div', attrib={'style': 'position:absolute;left:20px;top:200px'})
        description.append(div)

        h3 = ETree.Element('h3')
        h3.text = 'Metadata'
        div.append(h3)
        
        ul = ETree.Element('ul')
        div.append(ul)
        
        for text, value in self.metadata_fields(p).items():
            li = ETree.Element('li')
            li.text = text + str(value)
            ul.append(li)
        
        d = ETree.Element('div', attrib={'style': "position:absolute;left:300px;top:250px"})
        description.append(d)
        
        a = ETree.Element('a')
        if p.get('browse') is not None:
            a.set('href', p.get('browse')[0])
        else:
            a.set('href', "")
        
        d.append(a)
        
        img = ETree.Element('img')
        if p.get('thumbnailUrl') is not None:
            img.set('src', p.get('thumbnailUrl'))
        else:
            img.set('src', "None")
        a.append(img)
        
        styleUrl = ETree.Element('styleUrl')
        styleUrl.text = '#yellowLineGreenPoly'
        placemark.append(styleUrl)
        
        polygon = ETree.Element('Polygon')
        placemark.append(polygon)

        extrude = ETree.Element('extrude')
        extrude.text = '1'
        polygon.append(extrude)
        
        altitudeMode = ETree.Element('altitudeMode')
        altitudeMode.text = 'relativeToGround'
        polygon.append(altitudeMode)
        
        outerBondaryIs = ETree.Element('outerBoundaryIs')
        polygon.append(outerBondaryIs)
        
        linearRing = ETree.Element("LinearRing")
        outerBondaryIs.append(linearRing)
        
        coordinates = ETree.Element('coordinates')
        coordinates.text = '\n' + (14 * ' ') + ('\n' + (14 * ' ')).join([f"{c['Longitude']},{c['Latitude']},2000" for c in p['shape']]) + '\n' + (14 * ' ')
        linearRing.append(coordinates)

        self.indent(placemark, 3)
        
        # for CDATA section, manually replace &amp; escape character with &
        return ETree.tostring(placemark, encoding='unicode').replace('&amp;', '&')
    
    # Helper method for getting additional fields in <ul> tag
    def metadata_fields(self, item: dict):
        required = {
            'Processing type: ': item['processingTypeDisplay'],
            'Frame: ': item['frameNumber'],
            'Path: ': item['pathNumber'],
            'Orbit: ': item['orbit'],
            'Start time: ': item['startTime'],
            'End time: ': item['stopTime'],
        }
        
        optional = {}
        for text, key in [('Faraday Rotation: ', 'faradayRotation'), ('Ascending/Descending: ', 'flightDirection'), ('Off Nadir Angle: ', 'offNadirAngle'), ('Pointing Angle: ', 'pointingAngle'), ('Temporal Baseline: ', 'temporalBaseline'), ('Perpendicular Baseline: ', 'perpendicularBaseline')]:
            if item.get(key) is not None:
                if type(item[key]) == float and key == 'offNadirAngle':
                    optional[text] = f'{item[key]:g}' #trim trailing zeros    
                else:
                    optional[text] = item[key]
            elif key not in ['temporalBaseline', 'perpendicularBaseline']:
                optional[text] = 'None'
        
        output = { **required, **optional }
        if item['processingLevel'] == 'BURST':
            burst = {
                'Absolute Burst ID: ' :  item['burst']['absoluteBurstID'],
                'Relative Burst ID: ' :  item['burst']['relativeBurstID'],
                'Full Burst ID: ':  item['burst']['fullBurstID'],
                'Burst Index: ': item['burst']['burstIndex'],
                'Azimuth Time: ': item['burst']['azimuthTime'],
                'Azimuth Anx Time: ': item['burst']['azimuthAnxTime'],
                'Samples per Burst: ': item['burst']['samplesPerBurst'],
                'Subswath: ': item['burst']['subswath']
            }
            
            output = { **output, **burst}
        
        return output


from datetime import datetime
from typing import Any, Dict, List
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.constants import CMR_PAGE_SIZE
import re
from shapely import wkt
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry
from .field_map import field_map

import logging


def translate_opts(opts: ASFSearchOptions) -> list:
    # Need to add params which ASFSearchOptions cant support (like temporal),
    # so use a dict to avoid the validate_params logic:
    dict_opts = dict(opts)

    # Escape commas for each key in the list.
    # intersectsWith, temporal, and other keys you don't want to escape, so keep whitelist instead
    for escape_commas in ["campaign"]:
        if escape_commas in dict_opts:
            dict_opts[escape_commas] = dict_opts[escape_commas].replace(",", "\,")

    # Special case to unravel WKT field a little for compatibility
    if "intersectsWith" in dict_opts:
        shape = wkt.loads(dict_opts.pop('intersectsWith', None))

        # If a wide rectangle is provided, make sure to use the bounding box
        # instead of the wkt for better responses from CMR
        # This will provide better results with AOI's near poles
        if should_use_bbox(shape):
            bounds = shape.boundary.bounds
            if bounds[0] > 180 or bounds[2] > 180:
                bounds = [(x + 180) % 360 - 180 if idx % 2 == 0 and abs(x) > 180 else x for idx, x in enumerate(bounds)]

            bottom_left = [str(coord) for coord in bounds[:2]]
            top_right = [str(coord) for coord in bounds[2:]]

            bbox = ','.join([*bottom_left, *top_right])
            dict_opts['bbox'] = bbox
        else:
            (shapeType, shape) = wkt_to_cmr_shape(shape).split(':')
            dict_opts[shapeType] = shape

    # If you need to use the temporal key:
    if any(key in dict_opts for key in ['start', 'end', 'season']):
        dict_opts = fix_date(dict_opts)

    # convert the above parameters to a list of key/value tuples
    cmr_opts = []
    for (key, val) in dict_opts.items():
        # If it's "session" or something else CMR doesn't accept, don't send it:
        if key not in field_map:
            continue
        if isinstance(val, list):
            for x in val:
                if key in ['granule_list', 'product_list']:
                    for y in x.split(','):
                        cmr_opts.append((key, y))
                else:
                    if isinstance(x, tuple):
                        cmr_opts.append((key, ','.join([str(t) for t in x])))                    
                    else:
                        cmr_opts.append((key, x))
        else:
            cmr_opts.append((key, val))
    # translate the above tuples to CMR key/values
    for i, opt in enumerate(cmr_opts):
        cmr_opts[i] = field_map[opt[0]]['key'], field_map[opt[0]]['fmt'].format(opt[1])

    if should_use_asf_frame(cmr_opts):
            cmr_opts = use_asf_frame(cmr_opts)

    additional_keys = [
        ('page_size', CMR_PAGE_SIZE),
        ('options[temporal][and]', 'true'), 
        ('sort_key[]', '-end_date'), 
        ('sort_key[]', 'granule_ur'), 
        ('options[platform][ignore_case]', 'true'),
        ('provider', opts.provider),
    ]
    
    cmr_opts.extend(additional_keys)

    return cmr_opts


def should_use_asf_frame(cmr_opts):
    asf_frame_platforms = ['SENTINEL-1A', 'SENTINEL-1B', 'ALOS']

    return any([
        p[0] == 'platform[]' and p[1].upper() in asf_frame_platforms
        for p in cmr_opts
    ])


def use_asf_frame(cmr_opts):
    """
    Sentinel/ALOS: always use asf frame instead of esa frame

    Platform-specific hack
    We do them at the subquery level in case the main query crosses
    platforms that don't suffer these issue.
    """

    for n, p in enumerate(cmr_opts):
        if not isinstance(p[1], str):
            continue

        m = re.search(r'CENTER_ESA_FRAME', p[1])
        if m is None:
            continue

        logging.debug(
            'Sentinel/ALOS subquery, using ASF frame instead of ESA frame'
        )

        cmr_opts[n] = (
            p[0],
            p[1].replace(',CENTER_ESA_FRAME,', ',FRAME_NUMBER,')
        )
    
    return cmr_opts


def translate_product(item: dict) -> dict:
    try:
        coordinates = item['umm']['SpatialExtent']['HorizontalSpatialDomain']['Geometry']['GPolygons'][0]['Boundary']['Points']
        coordinates = [[c['Longitude'], c['Latitude']] for c in coordinates]
        geometry = {'coordinates': [coordinates], 'type': 'Polygon'}
    except KeyError as e:
        geometry = {'coordinates': None, 'type': 'Polygon'}

    umm = item.get('umm')

    properties = {
        'beamModeType': get(umm, 'AdditionalAttributes', ('Name', 'BEAM_MODE_TYPE'), 'Values', 0),
        'browse': get(umm, 'RelatedUrls', ('Type', [('GET RELATED VISUALIZATION', 'URL')])),
        'bytes': cast(int, try_round_float(get(umm, 'AdditionalAttributes', ('Name', 'BYTES'), 'Values', 0))),
        'centerLat': cast(float, get(umm, 'AdditionalAttributes', ('Name', 'CENTER_LAT'), 'Values', 0)),
        'centerLon': cast(float, get(umm, 'AdditionalAttributes', ('Name', 'CENTER_LON'), 'Values', 0)),
        'faradayRotation': cast(float, get(umm, 'AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0)),
        'fileID': get(umm, 'GranuleUR'),
        'flightDirection': get(umm, 'AdditionalAttributes', ('Name', 'ASCENDING_DESCENDING'), 'Values', 0),
        'groupID': get(umm, 'AdditionalAttributes', ('Name', 'GROUP_ID'), 'Values', 0),
        'granuleType': get(umm, 'AdditionalAttributes', ('Name', 'GRANULE_TYPE'), 'Values', 0),
        'insarStackId': get(umm, 'AdditionalAttributes', ('Name', 'INSAR_STACK_ID'), 'Values', 0),
        'md5sum': get(umm, 'AdditionalAttributes', ('Name', 'MD5SUM'), 'Values', 0),
        'offNadirAngle': cast(float, get(umm, 'AdditionalAttributes', ('Name', 'OFF_NADIR_ANGLE'), 'Values', 0)),
        'orbit': cast(int, get(umm, 'OrbitCalculatedSpatialDomains', 0, 'OrbitNumber')),
        'pathNumber': cast(int, get(umm, 'AdditionalAttributes', ('Name', 'PATH_NUMBER'), 'Values', 0)),
        'platform': get(umm, 'AdditionalAttributes', ('Name', 'ASF_PLATFORM'), 'Values', 0),
        'pointingAngle': cast(float, get(umm, 'AdditionalAttributes', ('Name', 'POINTING_ANGLE'), 'Values', 0)),
        'polarization': get(umm, 'AdditionalAttributes', ('Name', 'POLARIZATION'), 'Values', 0),
        'processingDate': get(umm, 'DataGranule', 'ProductionDateTime'),
        'processingLevel': get(umm, 'AdditionalAttributes', ('Name', 'PROCESSING_TYPE'), 'Values', 0),
        'sceneName': get(umm, 'DataGranule', 'Identifiers', ('IdentifierType', 'ProducerGranuleId'), 'Identifier'),
        'sensor': get(umm, 'Platforms', 0, 'Instruments', 0, 'ShortName'),
        'startTime': get(umm, 'TemporalExtent', 'RangeDateTime', 'BeginningDateTime'),
        'stopTime': get(umm, 'TemporalExtent', 'RangeDateTime', 'EndingDateTime'),
        'url': get(umm, 'RelatedUrls', ('Type', 'GET DATA'), 'URL'),
        'pgeVersion': get(umm, 'PGEVersionClass', 'PGEVersion')
    }
    
    if properties['beamModeType'] is None:
        properties['beamModeType'] = get(umm, 'AdditionalAttributes', ('Name', 'BEAM_MODE'), 'Values', 0)

    positions = {}
    velocities = {}
    positions['prePosition'], positions['prePositionTime'] = cast(get_state_vector, get(umm, 'AdditionalAttributes', ('Name', 'SV_POSITION_PRE'), 'Values', 0))
    positions['postPosition'], positions['postPositionTime'] = cast(get_state_vector, get(umm, 'AdditionalAttributes', ('Name', 'SV_POSITION_POST'), 'Values', 0))
    velocities['preVelocity'], velocities['preVelocityTime'] = cast(get_state_vector, get(umm, 'AdditionalAttributes', ('Name', 'SV_VELOCITY_PRE'), 'Values', 0))
    velocities['postVelocity'], velocities['postVelocityTime'] = cast(get_state_vector, get(umm, 'AdditionalAttributes', ('Name', 'SV_VELOCITY_POST'), 'Values', 0))
    ascendingNodeTime = get(umm, 'AdditionalAttributes', ('Name', 'ASC_NODE_TIME'), 'Values', 0)


    stateVectors = {
        'positions': positions,
        'velocities': velocities
    }

    insarBaseline = cast(float, get(umm, 'AdditionalAttributes', ('Name', 'INSAR_BASELINE'), 'Values', 0))
    
    baseline = {}
    if None not in stateVectors['positions'].values() and len(stateVectors.items()) > 0:
        baseline['stateVectors'] = stateVectors
        baseline['ascendingNodeTime'] = ascendingNodeTime
    elif insarBaseline is not None:
        baseline['insarBaseline'] = insarBaseline
    else:
        baseline = None

    if properties['url'] is not None:
        properties['fileName'] = properties['url'].split('/')[-1]
    else:
        properties['fileName'] = None

    if properties['platform'] is None:
        properties['platform'] = get(umm, 'Platforms', 0, 'ShortName')

    asf_frame_platforms = ['Sentinel-1A', 'Sentinel-1B', 'ALOS', 'SENTINEL-1A', 'SENTINEL-1B']
    if properties['platform'] in asf_frame_platforms:
        properties['frameNumber'] = cast(int, get(umm, 'AdditionalAttributes', ('Name', 'FRAME_NUMBER'), 'Values', 0))
    else:
        properties['frameNumber'] = cast(int, get(umm, 'AdditionalAttributes', ('Name', 'CENTER_ESA_FRAME'), 'Values', 0))

    if properties['processingLevel'] == 'BURST':
        burst = {
            'absoluteBurstID': cast(int, get(umm, 'AdditionalAttributes', ('Name', 'BURST_ID_ABSOLUTE'), 'Values', 0)),
            'relativeBurstID': cast(int, get(umm, 'AdditionalAttributes', ('Name', 'BURST_ID_RELATIVE'), 'Values', 0)),
            'fullBurstID': get(umm, 'AdditionalAttributes', ('Name', 'BURST_ID_FULL'), 'Values', 0),
            'burstIndex': cast(int, get(umm, 'AdditionalAttributes', ('Name', 'BURST_INDEX'), 'Values', 0)),
            'samplesPerBurst': cast(int, get(umm, 'AdditionalAttributes', ('Name', 'SAMPLES_PER_BURST'), 'Values', 0)),
            'subswath': get(umm, 'AdditionalAttributes', ('Name', 'SUBSWATH_NAME'), 'Values', 0),
            'azimuthTime': get(umm, 'AdditionalAttributes', ('Name', 'AZIMUTH_TIME'), 'Values', 0),
            'azimuthAnxTime': get(umm, 'AdditionalAttributes', ('Name', 'AZIMUTH_ANX_TIME'), 'Values', 0),
        }
        properties['burst'] = burst
        properties['sceneName'] = properties['fileID']
        properties['bytes'] = cast(int, get(umm, 'AdditionalAttributes', ('Name', 'BYTE_LENGTH'),  'Values', 0))

        urls = get(umm, 'RelatedUrls', ('Type', [('USE SERVICE API', 'URL')]), 0)
        if urls is not None:
            properties['url'] = urls[0]
            properties['fileName'] = properties['fileID'] + '.' + urls[0].split('.')[-1]
            properties['additionalUrls'] = [urls[1]]

    return {'geometry': geometry, 'properties': properties, 'type': 'Feature', 'baseline': baseline}

def get_additional_fields(umm, *field_path):
    return get(umm, *field_path)

def cast(f, v):
    try:
        return f(v)
    except TypeError:
        return None


def get(item: dict, *args):
    if item is None:
        return None
    for key in args:
        if isinstance(key, int):
            item = item[key] if key < len(item) else None
        elif isinstance(key, tuple):
            (a, b) = key
            if isinstance(b, List):
                output = []
                b = b[0]
                for child in item:
                    if get(child, key[0]) == b[0]:
                        output.append(get(child, b[1]))
                if len(output):
                    return output

                return None

            found = False
            for child in item:
                if get(child, a) == b:
                    item = child
                    found = True
                    break
            if not found:
                return None
        else:
            item = item.get(key)
        if item is None:
            return None
    if item in [None, 'NA', 'N/A', '']:
        item = None
    return item


def get_state_vector(state_vector: str):
    if state_vector is None:
        return None, None
    
    return list(map(float, state_vector.split(',')[:3])), state_vector.split(',')[-1]


# some products don't have integer values in BYTES fields, round to nearest int
def try_round_float(value: str):
    if value is not None:
        value = float(value)
        return round(value)
    
    return value


def fix_date(fixed_params: Dict[str, Any]):
    if 'start' in fixed_params or 'end' in fixed_params or 'season' in fixed_params:
        fixed_params["start"] = fixed_params["start"] if "start" in fixed_params else "1978-01-01T00:00:00Z"
        fixed_params["end"] = fixed_params["end"] if "end" in fixed_params else datetime.utcnow().isoformat()
        fixed_params["season"] = ','.join(str(x) for x in fixed_params['season']) if "season" in fixed_params else ""

        fixed_params['temporal'] = f'{fixed_params["start"]},{fixed_params["end"]},{fixed_params["season"]}'

        # And a little cleanup
        fixed_params.pop('start', None)
        fixed_params.pop('end', None)
        fixed_params.pop('season', None)
        
    return fixed_params


def should_use_bbox(shape: BaseGeometry):
    """
    If the passed shape is a polygon, and if that polygon
    is equivalent to it's bounding box (if it's a rectangle),
    we should use the bounding box to search instead
    """
    if isinstance(shape, Polygon):
        coords = [
            [shape.bounds[0], shape.bounds[1]], 
            [shape.bounds[2], shape.bounds[1]],
            [shape.bounds[2], shape.bounds[3]],
            [shape.bounds[0], shape.bounds[3]],
        ]
        return shape.equals(Polygon(shell=coords))
    
    return False


def wkt_to_cmr_shape(shape: BaseGeometry):
    # take note of the WKT type
    if shape.geom_type not in ["Point","LineString", "Polygon"]:
        raise ValueError('Unsupported WKT: {0}.'.format(shape.wkt))
    
    if shape.geom_type == "Polygon":
        coords = shape.exterior.coords
    else: # type == Point | Linestring
        coords = shape.coords
    # Turn [[x,y],[x,y]] into [x,y,x,y]:
    lon_lat_sequence = []
    for lon_lat in coords: lon_lat_sequence.extend(lon_lat)
    # Turn any "6e8" to a literal number. (As a sting):
    coords = ['{:.16f}'.format(float(cord)) for cord in lon_lat_sequence]
    return '{0}:{1}'.format(shape.geom_type.lower(), ','.join(coords))



from dateutil.parser import parse
import pytz
from .calc import calculate_perpendicular_baselines
from asf_search import ASFProduct, ASFSearchResults

precalc_datasets = ['AL', 'R1', 'E1', 'E2', 'J1']

def get_baseline_from_stack(reference: ASFProduct, stack: ASFSearchResults):
    warnings = None

    if len(stack) == 0:
        raise ValueError('No products found matching stack parameters')
    stack = [product for product in stack if not product.properties['processingLevel'].lower().startswith('metadata') and product.baseline != None]
    reference, stack, warnings = check_reference(reference, stack)
    
    stack = calculate_temporal_baselines(reference, stack)

    if get_platform(reference.properties['sceneName']) in precalc_datasets:
        stack = offset_perpendicular_baselines(reference, stack)
    else:
        stack = calculate_perpendicular_baselines(reference.properties['sceneName'], stack)

    return ASFSearchResults(stack), warnings

def valid_state_vectors(product: ASFProduct):
    if product is None:
        raise ValueError('Attempting to check state vectors on None, this is fatal')
    for key in ['postPosition', 'postPositionTime', 'prePosition', 'postPositionTime']:
        if key not in product.baseline['stateVectors']['positions'] or product.baseline['stateVectors']['positions'][key] == None:
            return False
    return True

def find_new_reference(stack: ASFSearchResults):
    for product in stack:
        if valid_state_vectors(product):
            return product
    return None

def check_reference(reference: ASFProduct, stack: ASFSearchResults):
    warnings = None
    if reference.properties['sceneName'] not in [product.properties['sceneName'] for product in stack]: # Somehow the reference we built the stack from is missing?! Just pick one
        reference = stack[0]
        warnings = [{'NEW_REFERENCE': 'A new reference scene had to be selected in order to calculate baseline values.'}]

    if get_platform(reference.properties['sceneName']) in precalc_datasets:
            if 'insarBaseline' not in reference.baseline:
                raise ValueError('No baseline values available for precalculated dataset')
    else:
        if not valid_state_vectors(reference): # the reference might be missing state vectors, pick a valid reference, replace above warning if it also happened
            reference = find_new_reference(stack)
            if reference == None:
                raise ValueError('No valid state vectors on any scenes in stack, this is fatal')
            warnings = [{'NEW_REFERENCE': 'A new reference had to be selected in order to calculate baseline values.'}]

    return reference, stack, warnings

def get_platform(reference: str):
    return reference[0:2].upper()

def get_default_product_type(product: ASFProduct):
    scene_name = product.properties['sceneName']
    
    if get_platform(scene_name) in ['AL']:
        return 'L1.1'
    if get_platform(scene_name) in ['R1', 'E1', 'E2', 'J1']:
        return 'L0'
    if get_platform(scene_name) in ['S1']:
        if product.properties['processingLevel'] == 'BURST':
            return 'BURST'
        return 'SLC'
    return None

def calculate_temporal_baselines(reference: ASFProduct, stack: ASFSearchResults):
    """
    Calculates temporal baselines for a stack of products based on a reference scene and injects those values into the stack.

    :param reference: The reference product from which to calculate temporal baselines.
    :param stack: The stack to operate on.
    :return: None, as the operation occurs in-place on the stack provided.
    """
    reference_time = parse(reference.properties['startTime'])
    if reference_time.tzinfo is None:
        reference_time = pytz.utc.localize(reference_time)

    for secondary in stack:
        secondary_time = parse(secondary.properties['startTime'])
        if secondary_time.tzinfo is None:
            secondary_time = pytz.utc.localize(secondary_time)
        secondary.properties['temporalBaseline'] = (secondary_time.date() - reference_time.date()).days
    
    return stack

def offset_perpendicular_baselines(reference: ASFProduct, stack: ASFSearchResults):
    reference_offset = float(reference.baseline['insarBaseline'])
    
    for product in stack:
            product.properties['perpendicularBaseline'] = round(float(product.baseline['insarBaseline']) - reference_offset)
    
    return stack



from typing import Iterable
from copy import copy

from asf_search.search import search
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.ASFSearchResults import ASFSearchResults


def product_search(
        product_list: Iterable[str],
        opts: ASFSearchOptions = None
) -> ASFSearchResults:
    """
    Performs a product ID search using the ASF SearchAPI

    :param product_list: List of specific products. Results guaranteed to be at most one product per product name.
    :param opts: An ASFSearchOptions object describing the search parameters to be used. Search parameters specified outside this object will override in event of a conflict.

    :return: ASFSearchResults(list) of search results
    """

    opts = (ASFSearchOptions() if opts is None else copy(opts))

    opts.merge_args(product_list=product_list)
    
    return search(opts=opts)



import unittest

from requests_mock import Adapter
import requests_mock
from asf_search.exceptions import ASFAuthenticationError, ASFDownloadError
import pytest
from unittest.mock import patch

import requests

from asf_search.download.download import download_url

def run_test_download_url_auth_error(url, path, filename):
    with patch('asf_search.ASFSession.get') as mock_get:
        resp = requests.Response()
        resp.status_code = 401
        mock_get.return_value = resp 
        
        if url == "pathError":
            with pytest.raises(ASFDownloadError):
                download_url(url, path, filename)

        with patch('asf_search.download.os.path.isdir') as path_mock:
            path_mock.return_value = True

            if url == "urlError":
                with patch('asf_search.download.os.path.isfile') as isfile_mock:
                    isfile_mock.return_value = False

                    with pytest.raises(ASFAuthenticationError):
                        download_url(url, path, filename)

            with patch('asf_search.download.os.path.isfile') as isfile_mock:
                isfile_mock.return_value = True
        
                with pytest.warns(Warning):
                    download_url(url, path, filename)

def run_test_download_url(url, path, filename):
    if filename == 'BURST':
        with patch('asf_search.ASFSession.get') as mock_get:
            resp = requests.Response()
            resp.status_code = 202
            resp.headers.update({'content-type': 'application/json'})
            mock_get.return_value = resp 

            with patch('asf_search.ASFSession.get') as mock_get_burst:
                resp_2 = requests.Response()
                resp_2.status_code = 200
                resp_2.headers.update({'content-type': 'image/tiff'})
                mock_get_burst.return_value = resp_2
                resp_2.iter_content = lambda chunk_size: []
                    
                with patch('builtins.open', unittest.mock.mock_open()) as m:
                    download_url(url, path, filename)
    else:
        with patch('asf_search.ASFSession.get') as mock_get:
            resp = requests.Response()
            resp.status_code = 200
            mock_get.return_value = resp
            resp.iter_content = lambda chunk_size: []
                
            with patch('builtins.open', unittest.mock.mock_open()) as m:
                download_url(url, path, filename)



import logging
from typing import Generator, Union, Iterable, Tuple
from copy import copy
from requests.exceptions import HTTPError
from requests import ReadTimeout, Response
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential
import datetime
import dateparser
import warnings

from asf_search import __version__

from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR import build_subqueries, translate_opts
from asf_search.ASFSession import ASFSession
from asf_search.ASFProduct import ASFProduct
from asf_search.exceptions import ASFSearch4xxError, ASFSearch5xxError, ASFSearchError
from asf_search.constants import INTERNAL
from asf_search.WKT.validate_wkt import validate_wkt
from asf_search.search.error_reporting import report_search_error


def search_generator(        
        absoluteOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        asfFrame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        beamMode: Union[str, Iterable[str]] = None,
        beamSwath: Union[str, Iterable[str]] = None,
        campaign: Union[str, Iterable[str]] = None,
        maxDoppler: float = None,
        minDoppler: float = None,
        end: Union[datetime.datetime, str] = None,
        maxFaradayRotation: float = None,
        minFaradayRotation: float = None,
        flightDirection: str = None,
        flightLine: str = None,
        frame: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        granule_list: Union[str, Iterable[str]] = None,
        groupID: Union[str, Iterable[str]] = None,
        insarStackId: str = None,
        instrument: Union[str, Iterable[str]] = None,
        intersectsWith: str = None,
        lookDirection: Union[str, Iterable[str]] = None,
        offNadirAngle: Union[float, Tuple[float, float], Iterable[Union[float, Tuple[float, float]]]] = None,
        platform: Union[str, Iterable[str]] = None,
        polarization: Union[str, Iterable[str]] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Union[str, Iterable[str]] = None,
        product_list: Union[str, Iterable[str]] = None,
        relativeOrbit: Union[int, Tuple[int, int], Iterable[Union[int, Tuple[int, int]]]] = None,
        season: Tuple[int, int] = None,
        start: Union[datetime.datetime, str] = None,
        absoluteBurstID: Union[int, Iterable[int]] = None,
        relativeBurstID: Union[int, Iterable[int]] = None,
        fullBurstID: Union[str, Iterable[str]] = None,
        collections: Union[str, Iterable[str]] = None,
        temporalBaselineDays: Union[str, Iterable[str]] = None,
        maxResults: int = None,
        opts: ASFSearchOptions = None,
        ) -> Generator[ASFSearchResults, None, None]:
    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = (ASFSearchOptions() if kwargs["opts"] is None else copy(opts))
    del kwargs["opts"]

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    maxResults = opts.pop("maxResults", None)

    if maxResults is not None and \
        (getattr(opts, 'granule_list', False) or getattr(opts, 'product_list', False)):
            raise ValueError("Cannot use maxResults along with product_list/granule_list.")

    preprocess_opts(opts)

    url = '/'.join(s.strip('/') for s in [f'https://{opts.host}', f'{INTERNAL.CMR_GRANULE_PATH}'])
    count = 0
    
    for query in build_subqueries(opts):
        translated_opts = translate_opts(query)
        try:
            response = get_page(session=opts.session, url=url, translated_opts=translated_opts, search_opts=query)
        except ASFSearchError as e:
            message = str(e)
            logging.error(message)
            report_search_error(query, message)
            opts.session.headers.pop('CMR-Search-After', None)
            return

        hits = [ASFProduct(f, session=query.session) for f in response.json()['items']]

        if 'CMR-Search-After' in response.headers:
            opts.session.headers.update({'CMR-Search-After': response.headers['CMR-Search-After']})

        if maxResults != None:
            last_page = ASFSearchResults(hits[:min(maxResults - count, len(hits))], opts=opts)
            count += len(last_page)
            
            if count == maxResults:
                last_page.searchComplete = True
                yield last_page
                return
            else:
                yield last_page
        else:
            count += len(hits)
            yield ASFSearchResults(hits, opts=opts)

        while('CMR-Search-After' in response.headers):
            opts.session.headers.update({'CMR-Search-After': response.headers['CMR-Search-After']})

            try:
                response = get_page(session=opts.session, url=url, translated_opts=translated_opts, search_opts=query)
            except ASFSearchError as e:
                message = str(e)
                logging.error(message)
                report_search_error(query, message)
                opts.session.headers.pop('CMR-Search-After', None)
                return

            hits = [ASFProduct(f, session=query.session) for f in response.json()['items']]

            if len(hits):
                if maxResults != None:
                    last_page = ASFSearchResults(hits[:min(maxResults - count, len(hits))], opts=opts)
                    count += len(last_page)
                    # results.extend(hits[:min(maxResults - len(results), len(hits))])
                    if count == maxResults:
                        last_page.searchComplete = True
                        yield last_page
                        return
                    else:
                        yield last_page
                else:
                    count += len(hits)
                    yield ASFSearchResults(hits, opts=opts)

        opts.session.headers.pop('CMR-Search-After', None)


@retry(reraise=True,
       retry=retry_if_exception_type((TimeoutError, ASFSearch5xxError)),
       wait=wait_exponential(multiplier=1, min=4, max=10),
       stop=stop_after_delay(340),
    )
def get_page(session: ASFSession, url: str, translated_opts: list, search_opts: ASFSearchOptions) -> Response:
    try:
        response = session.post(url=url, data=translated_opts, timeout=170)
        response.raise_for_status()
    except HTTPError as exc:
        error_message = f'HTTP {response.status_code}: {response.json()["errors"]}'
        if 400 <= response.status_code <= 499:
            raise ASFSearch4xxError(error_message) from exc
        if 500 <= response.status_code <= 599:
            raise ASFSearch5xxError(error_message) from exc
    except ReadTimeout as exc:
        raise ASFSearchError(f'Connection Error (Timeout): CMR took too long to respond ({url=})') from exc
    
    return response
    

def preprocess_opts(opts: ASFSearchOptions):
    # Repair WKT here so it only happens once, and you can save the result to the new Opts object:
    wrap_wkt(opts=opts)

    # Date/Time logic, convert "today" to the literal timestamp if needed:
    set_default_dates(opts=opts)

    # Platform Alias logic:
    set_platform_alias(opts=opts)


def wrap_wkt(opts: ASFSearchOptions):
    if opts.intersectsWith is not None:
        wrapped, _, __ = validate_wkt(opts.intersectsWith)
        opts.intersectsWith = wrapped.wkt


def set_default_dates(opts: ASFSearchOptions):
    if opts.start is not None and isinstance(opts.start, str):
        opts.start = dateparser.parse(opts.start, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    if opts.end is not None and isinstance(opts.end, str):
        opts.end = dateparser.parse(opts.end, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    # If both are used, make sure they're in the right order:
    if opts.start is not None and opts.end is not None:
        if opts.start > opts.end:
            warnings.warn(f"Start date ({opts.start}) is after end date ({opts.end}). Switching the two.")
            opts.start, opts.end = opts.end, opts.start
    # Can't do this sooner, since you need to compare start vs end:
    if opts.start is not None:
        opts.start = opts.start.strftime('%Y-%m-%dT%H:%M:%SZ')
    if opts.end is not None:
        opts.end = opts.end.strftime('%Y-%m-%dT%H:%M:%SZ')


def set_platform_alias(opts: ASFSearchOptions):
    # Platform Alias logic:
    if opts.platform is not None:
        plat_aliases = {
            # Groups:
            'S1': ['SENTINEL-1A', 'SENTINEL-1B'],
            'SENTINEL-1': ['SENTINEL-1A', 'SENTINEL-1B'],
            'SENTINEL': ['SENTINEL-1A', 'SENTINEL-1B'],
            'ERS': ['ERS-1', 'ERS-2'],
            'SIR-C': ['STS-59', 'STS-68'],
            # Singles / Aliases:
            'R1': ['RADARSAT-1'],
            'E1': ['ERS-1'],
            'E2': ['ERS-2'],
            'J1': ['JERS-1'],
            'A3': ['ALOS'],
            'AS': ['DC-8'],
            'AIRSAR': ['DC-8'],
            'SS': ['SEASAT 1'],
            'SEASAT': ['SEASAT 1'],
            'SA': ['SENTINEL-1A'],
            'SB': ['SENTINEL-1B'],
            'SP': ['SMAP'],
            'UA': ['G-III'],
            'UAVSAR': ['G-III'],
        }
        platform_list = []
        for plat in opts.platform:
            # If it's a key, *replace* it with all the values. Else just add the key:
            if plat.upper() in plat_aliases:
                platform_list.extend(plat_aliases[plat.upper()])
            else:
                platform_list.append(plat)

        opts.platform = list(set(platform_list))



from numbers import Number
import pytest
from typing import List

from shapely.wkt import loads
from shapely.geometry.base import BaseMultipartGeometry

from asf_search.WKT.validate_wkt import (
    validate_wkt,
    _search_wkt_prep,
    _get_clamped_and_wrapped_geometry, 
    _get_convex_hull, 
    _merge_overlapping_geometry, 
    _counter_clockwise_reorientation,
    _simplify_aoi,
    _get_shape_coords
)
from asf_search.exceptions import ASFWKTError


def run_test_validate_wkt_invalid_wkt_error(wkt: str):
    with pytest.raises(ASFWKTError):
        validate_wkt(wkt)


def run_test_validate_wkt_valid_wkt(wkt: str, validated_wkt: str):
    expected_aoi = loads(validated_wkt)
    actual_wrapped, actual_unwrapped, _ = validate_wkt(wkt)
    
    assert actual_wrapped.equals(expected_aoi), f"expected, {expected_aoi.wkt}, got {actual_wrapped.wkt}"
    
    actual_from_geom_wrapped, actual_from_geom_unwrapped, _ = validate_wkt(loads(wkt))
    assert actual_from_geom_wrapped.equals(expected_aoi)

def run_test_validate_wkt_clamp_geometry(wkt: str, clamped_wkt: str, clamped_count: Number, wrapped_count: Number):
    resp = _get_clamped_and_wrapped_geometry(loads(wkt))
    assert resp[0].wkt == clamped_wkt
    
    if clamped_count > 0:
        assert resp[2][0].report.split(' ')[2] == str(clamped_count)
    
    if wrapped_count > 0:
        assert resp[2][1].report.split(' ')[2] == str(wrapped_count)

def run_test_validate_wkt_convex_hull(wkt: str, corrected_wkt: str):
    shape = loads(wkt)
    assert corrected_wkt == _get_convex_hull(shape)[0].wkt 

def run_test_validate_wkt_merge_overlapping_geometry(wkt: str, merged_wkt: str):
    shape = loads(wkt)
    
    overlapping = _merge_overlapping_geometry(shape)
    if isinstance(overlapping, BaseMultipartGeometry):
        overlapping = overlapping.geoms
    assert overlapping[0].equals(loads(merged_wkt))

def run_test_validate_wkt_counter_clockwise_reorientation(wkt: str, cc_wkt: str):
    shape = loads(wkt)
    
    assert cc_wkt == _counter_clockwise_reorientation(shape)[0].wkt

def run_test_validate_wkt_get_shape_coords(wkt: str, coords: List[Number]):
    shape = loads(wkt)
    shape_coords = [[coord[0], coord[1]] for coord in _get_shape_coords(shape)]

    coords.sort()
    shape_coords.sort()

    assert len(shape_coords) == len(coords)
    assert shape_coords == coords

def run_test_search_wkt_prep(wkt: str):
    if wkt == ' ':
        with pytest.raises(ASFWKTError):
            _search_wkt_prep(None)
        
        return

    shape = loads(wkt)
    ls = _search_wkt_prep(shape)
    assert ls.geometryType() == shape.geometryType()
    assert shape.wkt == wkt
    
def run_test_simplify_aoi(wkt: str, simplified: str, repairs: List[str]):
    shape = loads(wkt)
    resp, shape_repairs = _simplify_aoi(shape)

    assert resp.equals(loads(simplified))

    for idx, repair in enumerate(repairs):
        assert shape_repairs[idx].report.startswith(repair)



from typing import List
from numbers import Number
from asf_search.baseline.stack import find_new_reference, get_baseline_from_stack, get_default_product_type, valid_state_vectors
from asf_search import ASFProduct, ASFSearchResults
import pytest
def run_test_find_new_reference(stack: List, output_index: Number) -> None:
    """
    Test asf_search.baseline.stack.find_new_reference
    """
    
    if stack == []:
        assert(find_new_reference(stack) == None)
    else:
        products = [ASFProduct(product) for product in stack]
        for idx, product in enumerate(products):
            product = clear_baseline(stack[idx], product)
        assert find_new_reference(products).properties['sceneName'] == stack[output_index]['properties']['sceneName']
        
def run_test_get_default_product_type(product: ASFProduct, product_type: str) -> None:
    assert get_default_product_type(product) == product_type
    
def run_test_get_baseline_from_stack(reference, stack, output_stack, error):
    reference = ASFProduct(reference)
    stack = ASFSearchResults([ASFProduct(product) for product in stack])
    
    if error == None:
        stack, warnings = get_baseline_from_stack(reference, stack)
        
        keys = ['sceneName', 'perpendicularBaseline', 'temporalBaseline']

        for idx, product in enumerate(stack):
            for key in keys:
                assert product.properties[key] == output_stack[idx]['properties'][key]
        
        return
    
    with pytest.raises(ValueError):
        for product in stack:
            if product.baseline.get('insarBaseline', False):
                product.baseline = {}
            else:
                product.baseline['stateVectors']['positions'] = {}
                product.baseline['stateVectors']['velocities'] = {}
        reference.baseline = {}
        get_baseline_from_stack(reference=reference, stack=stack)


def run_test_valid_state_vectors(reference, output):
    if reference != None:
        product = ASFProduct(reference)
        clear_baseline(reference, product)
        assert output == valid_state_vectors(product)
        return
    
    with pytest.raises(ValueError):
        valid_state_vectors(reference)

def clear_baseline(resource, product: ASFProduct):
# Baseline values can be restored from UMM in asfProduct constructor, 
# this erases them again if the resource omitted them from the product
    if resource['baseline'].get('stateVectors', False).get('positions', False) == {}:
        product.baseline['stateVectors']['positions'] = {}
        product.baseline['stateVectors']['velocities'] = {}
    
    return product



from asf_search.constants import INTERNAL
from asf_search.ASFSession import ASFSession

config = {
    'host': INTERNAL.CMR_HOST,
    'provider': INTERNAL.DEFAULT_PROVIDER,
    'session': ASFSession(),
}
